#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import unittest
import fwdpy11 as fp11


class testSlocusPop(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.pop = fp11.SlocusPop(1000)

    def test_N(self):
        self.assertEqual(self.pop.N, 1000)

    def test_generation(self):
        self.assertEqual(self.pop.generation, 0)

    def test_fitnesses(self):
        # All fitnesses should be 1
        self.assertEqual(
            sum([i.w for i in self.pop.diploids]),
            float(self.pop.N))

    def test_labels(self):
        # All diploids should be labeled 0 to pop.N-1
        self.assertEqual(
            [i.label for i in self.pop.diploids],
            [i for i in range(self.pop.N)])

    def test_genetic_values(self):
        self.assertEqual(sum([i.g for i in self.pop.diploids]), 0.0)

    def test_e_values(self):
        self.assertEqual(sum([i.e for i in self.pop.diploids]), 0.0)


class testSlocusPopExceptions(unittest.TestCase):
    def testNzero(self):
        with self.assertRaises(ValueError):
            fp11.SlocusPop(0)


class testSampling(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from quick_pops import quick_nonneutral_slocus
        self.pop = quick_nonneutral_slocus()
        self.rng = fp11.GSLrng(42)

    def testRandomSample(self):
        x = self.pop.sample(rng=self.rng, nsam=10)
        self.assertTrue(type(x) is tuple)
        x = self.pop.sample(rng=self.rng, nsam=10, separate=False)
        self.assertTrue(type(x) is list)
        x = self.pop.sample(rng=self.rng, nsam=10, remove_fixed=False)
        self.assertTrue(type(x) is tuple)
        x = self.pop.sample(rng=self.rng, nsam=10,
                            separate=True, remove_fixed=False)
        self.assertTrue(type(x) is tuple)

    def testDefinedSample(self):
        self.pop.sample(individuals=range(10))

        with self.assertRaises(IndexError):
            """
            fwdpp catches case where i >= N
            """
            self.pop.sample(individuals=range(self.pop.N, self.pop.N + 10))

        with self.assertRaises(Exception):
            """
            pybind11 disallows conversion of negative
            numbers to a list of unsigned types.
            """
            self.pop.sample(individuals=range(-10, 10))


class testPythonObjects(unittest.TestCase):
    @classmethod
    def setUp(self):
        from quick_pops import quick_slocus_qtrait_pop_params
        self.pop, self.pdict = quick_slocus_qtrait_pop_params()
        self.rng = fp11.GSLrng(101)

    def testInitialState(self):
        self.assertTrue(self.pop.popdata is None)
        self.assertTrue(self.pop.popdata_user is None)

    def testParentalData(self):
        from fwdpy11.model_params import SlocusParamsQ
        from fwdpy11.wright_fisher_qtrait import evolve
        params = SlocusParamsQ(**self.pdict)
        evolve(self.rng, self.pop, params)
        parents = [i.parental_data for i in self.pop.diploids]
        for i in parents:
            self.assertTrue(i is not None)
            self.assertTrue(len(i) == 2)
            self.assertTrue(i[0] < self.pop.N)
            self.assertTrue(i[1] < self.pop.N)

    def testPopdataReadOnly(self):
        from fwdpy11.model_params import SlocusParamsQ
        from fwdpy11.wright_fisher_qtrait import evolve
        params = SlocusParamsQ(**self.pdict)
        with self.assertRaises(Exception):
            class Recorder(object):
                def __call__(self, pop):
                    pop.popdata = []
            r = Recorder()
            evolve(self.rng, self.pop, params, r)

    def testPopdataUserAndPickling(self):
        from fwdpy11.model_params import SlocusParamsQ
        from fwdpy11.wright_fisher_qtrait import evolve
        params = SlocusParamsQ(**self.pdict)

        class Recorder(object):
            def __call__(self, pop):
                pop.popdata_user = pop.generation
        r = Recorder()
        evolve(self.rng, self.pop, params, r)
        self.assertEqual(self.pop.generation, self.pop.popdata_user)

        import pickle
        d = pickle.dumps(self.pop)
        up = pickle.loads(d)
        self.assertEqual(up.popdata_user, self.pop.generation)
        self.assertEqual(self.pop, up)


if __name__ == "__main__":
    unittest.main()



from .wfevolve_qtrait import evolve_singlepop_regions_qtrait_cpp
from .wfevolve_qtrait import evolve_qtrait_mloc_regions_cpp
import fwdpy11.gsl_random
import math

# Note: making changes until "END LINE NUMBER EMBARGO"
# is seen below requires changes to the documentation as
# well!  The fwdpy11 manual includes specific line ranges
# from this file in order to provide examples of trait ->
# fitness and noise functions.  The relevant manual file
# is doc/examples/qtraits.rst


class GSS:
    """
    Gaussian stabilizing selection with a constant optimum.

    This is a callable object.  The __call__ function
    takes a trait value and return a fitness.

    Its parameters are given below.

    :param trait_value: A trait (phenotype) value

    :return: :math:`w=e^{-\\frac{(G-E)^2}{2VS}}`

    :rtype: float
    """

    def __init__(self, VS, O):
        """
        :param VS: 1/VS is intensity of selection against
            phenotypic deviations from the mean/optimum.
        :param O: The optimum trait value.
        """
        if VS <= 0.:
            raise ValueError("VS > 0 required")
        self.VS = VS
        self.O = O

    def __call__(self, g, e):
        devsq = pow((g + e) - self.O, 2)
        return math.exp(-devsq / (2.0 * self.VS))


class GSSmo:
    """
    Gaussian stabilizing selection with moving optimum.

    .. note::
        This class's constructor enforces nothing about the initial
        values for the model parameters. The reason is for flexibility,
        in that you may run a simulation, and then evolve more later
        one.  However, it makes sense that, when you start a simulation,
        the first tuple for the constructor refers to the population's
        current generation.
    """

    def __init__(self, optima):
        """
        :param optima: A list of tuples.  Each tuple is (generation,optimum,VS)
        """
        if len(optima) == 0:
            raise ValueError("empty list of optima")
        for oi in optima:
            if isinstance(oi, tuple) is False:
                raise ValueError("optima must cointain tuples")
            if len(oi) != 3:
                raise ValueError("incorrect tuple length")
            if oi[0] < 0:
                raise ValueError("negative generation not allowed")
            if oi[2] <= 0.0:
                raise ValueError("VS > 0 required")
        self.optima = optima.copy()
        self.env = self.optima.pop(0)

    def __call__(self, g, e):
        devsq = pow((g + e) - self.env[1], 2)
        return math.exp(-devsq / (2.0 * self.env[2]))

    def update(self, pop):
        """
        Update the fitness model conditions.

        :param generation: the generation in the simulation

        .. note:: this function is called from within the simulation
        """
        if len(self.optima) == 0:
            return
        if pop.generation >= self.optima[0][0]:
            self.env = self.optima.pop(0)


class GaussianNoise:
    """
    Gaussian noise for trait values.

    Adds :math:`N(\\mu,\\sigma)` to trait values.
    """

    def __init__(self, rng, sd, mean=0.0):
        """
        :param rng: A :class:`fwdpy11.GSLrng`
        :param sd: :math:`\\sigma`
        :param mean: (0.0) :math:`\\mu`
        """
        if(type(rng) is fwdpy11.GSLrng) is False:
            raise ValueError("rng must be a fwdpy11.GSLrng")
        self.sd = sd
        self.mean = mean
        self.rng = rng

    def __call__(self, parent1, parent2):
        return self.mean + \
            fwdpy11.gsl_random.gsl_ran_gaussian_ziggurat(self.rng, self.sd)

# END LINE NUMBER EMBARGO##


def _evolve_slocus(rng, pop, params, recorder=None):
    import warnings
    # Test parameters while suppressing warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        params.validate()
    from .internal import makeMutationRegions, makeRecombinationRegions
    from functools import partial
    mm = makeMutationRegions(params.nregions, params.sregions)
    rm = makeRecombinationRegions(params.recregions)

    noise = None
    if params.noise is None:
        noise = GaussianNoise(rng, 0.)
    else:
        noise = params.noise
    updater = None
    noise_updater = None
    if hasattr(params.trait2w, 'update'):
        updater = partial(type(params.trait2w).update, params.trait2w)
    if hasattr(noise, 'update'):
        noise_updater = partial(type(noise).update, noise)
    if recorder is None:
        from fwdpy11.temporal_samplers import RecordNothing
        recorder = RecordNothing()

    prune_selected = None
    if params.prune_selected is not None:
        prune_selected = params.prune_selected
    evolve_singlepop_regions_qtrait_cpp(rng, pop, params.demography,
                                        params.mutrate_n, params.mutrate_s,
                                        params.recrate, mm, rm,
                                        params.gvalue, recorder,
                                        params.pself, params.trait2w, updater,
                                        noise, noise_updater, prune_selected)


def _evolve_mlocus(rng, pop, params, recorder=None):
    import warnings
    # Test parameters while suppressing warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        params.validate()

    from .internal import makeMutationRegions, makeRecombinationRegions
    from functools import partial

    noise = None
    if params.noise is None:
        noise = GaussianNoise(rng, 0.)
    else:
        noise = params.noise
    mm = [makeMutationRegions(i, j) for i, j in zip(params.nregions,
                                                    params.sregions)]
    rm = [makeRecombinationRegions(i) for i in params.recregions]
    updater = None
    noise_updater = None
    if hasattr(params.trait2w, 'update'):
        updater = partial(type(params.trait2w).update, params.trait2w)
    if hasattr(noise, 'update'):
        noise_updater = partial(type(noise).update, noise)
    if recorder is None:
        from fwdpy11.temporal_samplers import RecordNothing
        recorder = RecordNothing()
    prune_selected = None
    if params.prune_selected is not None:
        prune_selected = params.prune_selected
    else:
        prune_selected = False
    if prune_selected is True:
        warnings.warn(
            "selected fixations will be removed from population during the simulation")
    evolve_qtrait_mloc_regions_cpp(rng, pop, params.demography,
                                   params.mutrates_n, params.mutrates_s,
                                   params.recrates, mm, rm, params.interlocus,
                                   params.gvalue,
                                   recorder,
                                   params.pself,
                                   params.aggregator,
                                   params.trait2w,
                                   updater, noise, noise_updater,
                                   prune_selected)


def evolve(rng, pop, params, recorder=None):
    """
    Evolve a quantitative trait.

    This function evolves a class of models where traits are calculated
    and then mapped to fitness.  See the sections of the manual on
    simulating quantitative traits.

    :param rng: An instance of :class:`fwdpy11.GSLrng`.
    :param pop: An instance of :class:`fwdpy11.Spop` or :class:`fwdpy11.MlocusPop`.
    :param params: An instance of :class:`fwdpy11.model_params.SlocusParamsQ`
        or :class:`fwdpy11.model_params.MlocusParamsQ`.
    :param recorder: (None) A callable to record data from the population.

    .. note::
        If recorder is None,
        :class:`fwdpy11.temporal_samplers.RecordNothing` will be used.

    .. note::
        If params.noise is None,
        :class:`fwdpy11.wright_fisher_qtrait.GaussianNoise` will be used
        with mean and standard deviation both set to zero.

    .. note::
        Please be sure to match your population type
        to your model parameter type.  For example,
        mixing a single-locus population with
        multi-locus model parameters will trigger an exception.
    """
    import warnings
    # Test parameters while suppressing warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        params.validate()

    if recorder is None:
        from fwdpy11.temporal_samplers import RecordNothing
        recorder = RecordNothing()
    try:
        _evolve_slocus(rng, pop, params, recorder)
    except:
        _evolve_mlocus(rng, pop, params, recorder)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

# TYPETEST

import fwdpy11
import unittest


class testSlocusPopCreate(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.mutations = fwdpy11.VecMutation()
        self.fixations = fwdpy11.VecMutation()
        self.gametes = fwdpy11.VecGamete()
        self.diploids = fwdpy11.VecDiploid()
        self.mutations.append(fwdpy11.Mutation(0.1, -0.01, 1.0, 0, 0))
        self.fixations.append(fwdpy11.Mutation(0.1, -0.01, 1.0, 0, 0))
        self.gametes.append(fwdpy11.Gamete(
            (2, fwdpy11.VecUint32([]), fwdpy11.VecUint32([0]))))
        self.diploids.append(fwdpy11.SingleLocusDiploid(0, 0))

    def testConstruction(self):
        pop = fwdpy11.SlocusPop(self.diploids, self.gametes, self.mutations)
        self.assertTrue(pop.N, 1)

    def testStaticMethod(self):
        pop = fwdpy11.SlocusPop.create(
            self.diploids, self.gametes, self.mutations)
        self.assertTrue(type(pop) is fwdpy11.SlocusPop)
        # Test that data were moved and not copied:
        self.assertEqual(len(self.diploids), 0)
        self.assertEqual(len(self.gametes), 0)
        self.assertEqual(len(self.mutations), 0)

    def testStaticMethodWithFixations(self):
        ftimes = fwdpy11.VecUint32([1])
        pop = fwdpy11.SlocusPop.create(self.diploids,
                                       self.gametes,
                                       self.mutations,
                                       self.fixations,
                                       ftimes, 2)
        self.assertTrue(type(pop) is fwdpy11.SlocusPop)
        self.assertEqual(len(pop.fixations), len(pop.fixation_times))
        # Test that data were moved and not copied:
        self.assertEqual(len(self.diploids), 0)
        self.assertEqual(len(self.gametes), 0)
        self.assertEqual(len(self.mutations), 0)
        self.assertEqual(len(self.fixations), 0)
        self.assertEqual(len(ftimes), 0)


# class testSlocusPopGeneralMutVecCreate(unittest.TestCase):
#     @classmethod
#     def setUp(self):
#         self.mutations = fwdpy11.VecGeneralMutVec()
#         self.fixations = fwdpy11.VecGeneralMutVec()
#         self.gametes = fwdpy11.VecGamete()
#         self.diploids = fwdpy11.VecDiploid()
#         s = fwdpy11.VecDouble([-0.1, 0.1])
#         h = fwdpy11.VecDouble([0.0, 1.0])
#         m = fwdpy11.GeneralMutVec((s, h, 0.1, 3, 0))
#         self.mutations.append(m)
#         self.fixations.append(m)
#         self.gametes.append(fwdpy11.Gamete(
#             (2, fwdpy11.VecUint32([]), fwdpy11.VecUint32([0]))))
#         self.diploids.append(fwdpy11.SingleLocusDiploid(0, 0))
# 
#     def testConstruction(self):
#         pop = fwdpy11.SlocusPopGeneralMutVec(
#             self.diploids, self.gametes, self.mutations)
#         self.assertTrue(pop.N, 1)
# 
#     def testStaticMethod(self):
#         pop = fwdpy11.SlocusPopGeneralMutVec.create(
#             self.diploids, self.gametes, self.mutations)
#         self.assertTrue(type(pop) is fwdpy11.SlocusPopGeneralMutVec)
#         # Test that data were moved and not copied:
#         self.assertEqual(len(self.diploids), 0)
#         self.assertEqual(len(self.gametes), 0)
#         self.assertEqual(len(self.mutations), 0)
# 
#     def testStaticMethodWithFixations(self):
#         ftimes = fwdpy11.VecUint32([1])
#         pop = fwdpy11.SlocusPopGeneralMutVec.create(self.diploids,
#                                                     self.gametes,
#                                                     self.mutations,
#                                                     self.fixations,
#                                                     ftimes, 2)
#         self.assertTrue(type(pop) is fwdpy11.SlocusPopGeneralMutVec)
#         self.assertEqual(len(pop.fixations), len(pop.fixation_times))
#         # Test that data were moved and not copied:
#         self.assertEqual(len(self.diploids), 0)
#         self.assertEqual(len(self.gametes), 0)
#         self.assertEqual(len(self.mutations), 0)
#         self.assertEqual(len(self.fixations), 0)
#         self.assertEqual(len(ftimes), 0)


class testMlocusPopCreate(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.mutations = fwdpy11.VecMutation()
        self.fixations = fwdpy11.VecMutation()
        self.gametes = fwdpy11.VecGamete()
        self.diploids = fwdpy11.VecVecDiploid()
        self.mutations.append(fwdpy11.Mutation(0.1, -0.01, 1.0, 0, 0))
        self.fixations.append(fwdpy11.Mutation(0.1, -0.01, 1.0, 0, 0))
        self.gametes.append(fwdpy11.Gamete(
            (4, fwdpy11.VecUint32([]), fwdpy11.VecUint32([0]))))
        self.diploids.append(fwdpy11.VecDiploid(
            [fwdpy11.SingleLocusDiploid(0, 0)] * 2))

    def testConstruction(self):
        pop = fwdpy11.MlocusPop(self.diploids, self.gametes, self.mutations)
        self.assertTrue(pop.N, 1)

    def testStaticMethod(self):
        pop = fwdpy11.MlocusPop.create(
            self.diploids, self.gametes, self.mutations)
        self.assertTrue(type(pop) is fwdpy11.MlocusPop)
        # Test that data were moved and not copied:
        self.assertEqual(len(self.diploids), 0)
        self.assertEqual(len(self.gametes), 0)
        self.assertEqual(len(self.mutations), 0)

    def testStaticMethodWithFixations(self):
        ftimes = fwdpy11.VecUint32([1])
        pop = fwdpy11.MlocusPop.create(self.diploids,
                                       self.gametes,
                                       self.mutations,
                                       self.fixations,
                                       ftimes, 2)
        self.assertTrue(type(pop) is fwdpy11.MlocusPop)
        self.assertEqual(len(pop.fixations), len(pop.fixation_times))
        # Test that data were moved and not copied:
        self.assertEqual(len(self.diploids), 0)
        self.assertEqual(len(self.gametes), 0)
        self.assertEqual(len(self.mutations), 0)
        self.assertEqual(len(self.fixations), 0)
        self.assertEqual(len(ftimes), 0)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
from .wfevolve import evolve_singlepop_regions_cpp


def evolve(rng, pop, params, recorder=None):
    """
    Evolve a population

    :param rng: An instance of :class:`fwdpy11.GSLrng`
    :param pop: An instance of :class:`fwdpy11.SlocusPop`
    :param params: An instance of :class:`fwdpy11.model_params.SlocusParams`
    :param recorder: (None) A temporal sampler/data recorder.

    .. note::
        If recorder is None,
        then :class:`fwdpy11.temporal_samplers.RecordNothing` will be used.

    """
    import warnings
    # Test parameters while suppressing warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # Will throw exception if anything is wrong:
        params.validate()

    from .internal import makeMutationRegions, makeRecombinationRegions
    mm = makeMutationRegions(params.nregions, params.sregions)
    rm = makeRecombinationRegions(params.recregions)

    if recorder is None:
        from fwdpy11.temporal_samplers import RecordNothing
        recorder = RecordNothing()

    evolve_singlepop_regions_cpp(rng, pop, params.demography,
                                 params.mutrate_n, params.mutrate_s,
                                 params.recrate, mm, rm,
                                 params.gvalue, recorder, params.pself,
                                 params.prune_selected)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import unittest

import fwdpy11 as fp11
import fwdpy11.model_params as fp11mp
import fwdpy11.multilocus as ml
import fwdpy11.wright_fisher_qtrait as wfq
from fwdpy11.fitness import SlocusAdditive
from quick_pops import quick_slocus_qtrait_pop_params
import numpy as np
import pickle


class testSlocusParams(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # We do not assign certain attributes,
        # which allows us to test the defaults.
        self.pdict = {'nregions': [],
                      'sregions': [],
                      'recregions': [],
                      'rates': [0, 0, 0],
                      'demography': np.array([100] * 100, dtype=np.uint32)
                      }

    def test_SlocusParamsDefaults(self):
        from fwdpy11.fitness import SlocusMult
        m = fp11mp.SlocusParams(**self.pdict)
        self.assertEqual(type(m.gvalue), SlocusMult)

    def test_SlocusParamsQDefaults(self):
        self.pdict['prune_selected'] = False  # Suppresses a warning
        from fwdpy11.trait_values import SlocusAdditiveTrait
        from fwdpy11.wright_fisher_qtrait import GSS
        m = fp11mp.SlocusParamsQ(**self.pdict)
        self.assertEqual(type(m.gvalue), SlocusAdditiveTrait)
        self.assertEqual(type(m.trait2w), GSS)
        self.assertEqual(m.trait2w.VS, 1.0)
        self.assertEqual(m.trait2w.O, 0.0)

    def test_set_nregions(self):
        self.setUpClass()
        m = fp11mp.SlocusParams(**self.pdict)
        try:
            m.nregions = [fp11.Region(0, 1, 1)]
        except:
            self.fail("unexpected exception")
        # An empty list is ok
        try:
            m.nregions = []
        except:
            self.fail("unexpected exception")
        # TypeError b/c not iterable:
        with self.assertRaises(TypeError):
            m.nregions = fp11.Region(0, 1, 1)
        with self.assertRaises(TypeError):
            m.nregions = [fp11.ExpS(0, 1, 1, 0.25)]

    def test_set_recregions(self):
        self.setUpClass()
        m = fp11mp.SlocusParams(**self.pdict)
        try:
            m.recregions = [fp11.Region(0, 1, 1)]
        except:
            self.fail("unexpected exception")
        # An empty list is ok
        try:
            m.recregions = []
        except:
            self.fail("unexpected exception")
        # TypeError b/c not iterable:
        with self.assertRaises(TypeError):
            m.recregions = fp11.Region(0, 1, 1)
        with self.assertRaises(TypeError):
            m.recregions = [fp11.ExpS(0, 1, 1, 0.25)]

    def test_set_sregions(self):
        self.setUpClass()
        m = fp11mp.SlocusParams(**self.pdict)
        try:
            m.sregions = [fp11.GaussianS(0, 1, 1, 0.1)]
        except:
            self.fail("unexpected exception")
        # An empty list is ok
        try:
            m.sregions = []
        except:
            self.fail("unexpected exception")
        with self.assertRaises(TypeError):
            m.sregions = fp11.GaussianS(0, 1, 1, 0.2)
        with self.assertRaises(TypeError):
            m.sregions = [fp11.Region(0, 1, 1)]

    def test_init_demog(self):
        self.setUpClass()
        try:
            m = fp11mp.SlocusParams(**self.pdict)
            m
        except:
            self.fail("unexpected exception")

    def test_init_demog_list(self):
        self.pdict['demography'] = [100] * 10
        with self.assertRaises(ValueError):
            m = fp11mp.SlocusParams(**self.pdict)
            m

    def test_init_bad_popsizes(self):
        self.pdict['demography'] = np.array([100] * 10 + [-1])
        with self.assertRaises(ValueError):
            m = fp11mp.SlocusParams(**self.pdict)
            m

    def test_bad_rates(self):
        self.setUpClass()
        self.pdict['mutrate_n'] = None
        with self.assertRaises(ValueError):
            m = fp11mp.SlocusParams(**self.pdict)
            m
        self.pdict['mutrate_n'] = -0.01
        with self.assertRaises(ValueError):
            m = fp11mp.SlocusParams(**self.pdict)
            m

    def test_warning_for_zero_rate(self):
        # We pass in some sregions but leave mutrate_s == 0.
        self.setUpClass()
        self.pdict['sregions'] = [fp11.GammaS(0, 1, 1, -0.1, 0.10, 1.0)]
        with self.assertWarns(UserWarning):
            m = fp11mp.SlocusParams(**self.pdict)
            m.validate()

    def test_set_rates_list(self):
        self.setUpClass()
        self.pdict['rates'] = [0., 1e-3, 0.25]
        try:
            m = fp11mp.SlocusParams(**self.pdict)
            m
        except:
            self.fail("unexpected exception")

    def test_set_rates_tuple(self):
        m = fp11mp.SlocusParams()
        m.rates = (0., 1e-3, 0.25)
        self.assertEqual(m.mutrate_n, 0.0)
        self.assertEqual(m.mutrate_s, 1e-3)
        self.assertEqual(m.recrate, 0.25)
        with self.assertRaises(ValueError):
            m.rates = (0., -1e-3, 0.25)

    def test_set_rates_dict(self):
        m = fp11mp.SlocusParams()
        with self.assertRaises(ValueError):
            m.rates = {'mutrate_n': 1e-3}
        with self.assertRaises(ValueError):
            m.rates = {'mutrate_n': -1e-3,
                       'mutrate_s': 1e-3,
                       'recrate': 1e-3}
        with self.assertRaises(ValueError):
            m.rates = {'monkeys_n': 1e-3}

    def test_set_gvalue(self):
        m = fp11mp.SlocusParams()
        from fwdpy11.fitness import SlocusAdditive
        m.gvalue = SlocusAdditive()

    def test_set_bad_gvalue(self):
        m = fp11mp.SlocusParams()
        with self.assertRaises(ValueError):
            m.gvalue = []

    def test_invalid_kwargs(self):
        from fwdpy11.fitness import SlocusAdditive
        with self.assertRaises(ValueError):
            m = fp11mp.SlocusParams(fitness=SlocusAdditive())
            m


class testMlocusParams(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.rng = fp11.GSLrng(42)
        nregions = [[fp11.Region(0, 1, 1)], [fp11.Region(1, 2, 1)]]
        sregions = [
            [fp11.ExpS(0, 1, 1, -0.1)],
            [fp11.GammaS(1, 2, 1, -0.1, 0.5)]]
        recregions = nregions
        interlocus = ml.binomial_rec([0.5])
        region_rates = [1e-3, 1e-3]
        genetic_value = ml.MultiLocusGeneticValue([SlocusAdditive()] * 2)
        nlist = np.array([100] * 10, dtype=np.uint32)
        self.param_dict = {'nregions': nregions,
                           'sregions': sregions,
                           'recregions': recregions,
                           'interlocus': interlocus,
                           'mutrates_n': region_rates,
                           'mutrates_s': region_rates,
                           'recrates': region_rates,
                           'gvalue': genetic_value,
                           'demography': nlist}
        self.m = fp11mp.MlocusParams(**self.param_dict)

    def test_assign_gvalue(self):
        self.m.gvalue = ml.MultiLocusGeneticValue([SlocusAdditive()] * 2)

    def test_assign_gvalue_from_list(self):
        self.m.gvalue = [SlocusAdditive()] * 2

    def test_assign_gvalue_from_bad_input(self):
        with self.assertRaises(ValueError):
            self.m.gvalue = [2.0] * 2

    def test_assign_aggregator(self):
        self.m.aggregator = ml.AggAddTrait()

    def test_assign_aggregator_non_callable(self):
        with self.assertRaises(ValueError):
            self.m.aggregator = 2.0
    # Now, tests that check the validate(), too

    def test_construction(self):
        self.setUpClass()
        try:
            self.m.validate()
        except:
            self.fail("unexpected exception")

    def test_construction_unequal_region_lens_1(self):
        self.param_dict['nregions'] = [fp11.Region(0, 1, 1)]
        with self.assertRaises(TypeError):
            self.m = fp11mp.MlocusParams(**self.param_dict)

    def test_construction_unequal_region_lens_2(self):
        self.param_dict['sregions'] = [fp11.ExpS(0, 1, 1, 0.25)]
        with self.assertRaises(TypeError):
            self.m = fp11mp.MlocusParams(**self.param_dict)

    def test_construction_unequal_region_lens_3(self):
        self.param_dict['recregions'] = [fp11.Region(0, 1, 1)]
        with self.assertRaises(TypeError):
            self.m = fp11mp.MlocusParams(**self.param_dict)

    def test_empty_sregion_exception(self):
        self.setUpClass()
        self.param_dict['sregions'] = [[fp11.ExpS(0, 1, 1, 0.25)], []]
        self.m = fp11mp.MlocusParams(**self.param_dict)
        with self.assertRaises(ValueError):
            self.m.validate()

    def test_empty_nregion_exception(self):
        self.setUpClass()
        self.param_dict['nregions'] = [[], []]
        self.m = fp11mp.MlocusParams(**self.param_dict)
        with self.assertRaises(ValueError):
            self.m.validate()

    def test_empty_recregion_exception(self):
        self.setUpClass()
        self.param_dict['recregions'] = [[fp11.Region(0, 1, 1)], []]
        self.m = fp11mp.MlocusParams(**self.param_dict)
        with self.assertRaises(ValueError):
            self.m.validate()

    def test_pickle_pdict(self):
        self.setUpClass()
        p = pickle.dumps(self.param_dict)
        up = pickle.loads(p)
        up

    def test_pickle_params(self):
        self.setUpClass()
        self.m = fp11mp.MlocusParams(**self.param_dict)
        p = pickle.dumps(self.m)
        up = pickle.loads(p)
        up

    def test_deprecated_keyword(self):
        self.setUpClass()
        self.m = fp11mp.MlocusParams(**self.param_dict)
        with self.assertWarns(DeprecationWarning):
            self.m.agg = ml.AggAddTrait()


class testMlocusParamsDefaults(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        nregions = [[fp11.Region(0, 1, 1)], [fp11.Region(1, 2, 1)]]
        sregions = [
            [fp11.ExpS(0, 1, 1, -0.1)],
            [fp11.GammaS(1, 2, 1, -0.1, 0.5)]]
        recregions = nregions
        interlocus = ml.binomial_rec([0.5])
        region_rates = [1e-3, 1e-3]
        nlist = np.array([100] * 10, dtype=np.uint32)
        self.pdict = {'nregions': nregions,
                      'sregions': sregions,
                      'recregions': recregions,
                      'interlocus': interlocus,
                      'mutrates_n': region_rates,
                      'mutrates_s': region_rates,
                      'recrates': region_rates,
                      'demography': nlist
                      }

    def test_MlocusParamsDefaults(self):
        m = fp11mp.MlocusParams(**self.pdict)
        from fwdpy11.multilocus import AggMultFitness
        from fwdpy11.multilocus import MultiLocusGeneticValue
        from fwdpy11.fitness import SlocusMult
        self.assertEqual(type(m.aggregator), AggMultFitness)
        self.assertEqual(type(m.gvalue), MultiLocusGeneticValue)
        for i in m.gvalue.fitness_functions:
            self.assertEqual(type(i), SlocusMult)

    def test_MlocusParamsQDefaults(self):
        self.pdict['prune_selected'] = True
        m = fp11mp.MlocusParamsQ(**self.pdict)
        from fwdpy11.multilocus import AggMultTrait
        from fwdpy11.multilocus import MultiLocusGeneticValue
        from fwdpy11.trait_values import SlocusAdditiveTrait
        from fwdpy11.wright_fisher_qtrait import GSS
        self.assertEqual(type(m.aggregator), AggMultTrait)
        self.assertEqual(type(m.gvalue), MultiLocusGeneticValue)
        self.assertEqual(type(m.trait2w), GSS)
        self.assertEqual(m.trait2w.VS, 1.0)
        self.assertEqual(m.trait2w.O, 0.0)
        for i in m.gvalue.fitness_functions:
            self.assertEqual(type(i), SlocusAdditiveTrait)

    def test_MlocusParams_cannot_set_gvalue(self):
        self.pdict.pop('sregions')
        m = fp11mp.MlocusParams(**self.pdict)
        self.assertIs(m.gvalue, None)
        m = fp11mp.MlocusParamsQ(**self.pdict)
        self.assertIs(m.gvalue, None)


class test_SlocusParamsQEvolve(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.pop, self.pdict = quick_slocus_qtrait_pop_params()
        self.params = fp11mp.SlocusParamsQ(**self.pdict)
    def test_evolve(self):
        rng = fp11.GSLrng(42)
        wfq.evolve(rng, self.pop, self.params)


if __name__ == "__main__":
    unittest.main()



import unittest
import fwdpy11 as fp11
import fwdpy11.multilocus as fp11m
import fwdpy11.fitness as fp11w
import numpy as np
import pickle

class testMultiLocusFitness(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.pop = fp11.MlocusPop(1000, 5)
        self.all_additive = [fp11w.SlocusAdditive()] * 5
        self.all_mult = [fp11w.SlocusMult()] * 5
        self.agg_add_w = fp11m.AggAddFitness()
        self.agg_add_t = fp11m.AggAddTrait()
        self.agg_mult_w = fp11m.AggMultFitness()
        self.agg_mult_t = fp11m.AggMultTrait()
        self.additive_w = fp11m.MultiLocusGeneticValue(self.all_additive)
        self.mult_w = fp11m.MultiLocusGeneticValue(self.all_mult)
        self.additive_fitness_value_for_dip = self.additive_w(
            self.pop.diploids[0], self.pop)
        self.mult_fitness_value_for_dip = self.mult_w(
            self.pop.diploids[0], self.pop)

    def add_it_up(self, x):
        # naive/incorrect aggregator,
        # but serves to test type(x)
        self.assertEqual(isinstance(x, np.ndarray), True)
        return x.sum()

    def prod_it_up(self, x):
        # naive/incorrect aggregator,
        # but serves to test type(x)
        self.assertEqual(isinstance(x, np.ndarray), True)
        return x.prod()

    def testAllAdditive(self):
        self.assertEqual(len(self.all_additive), 5)
        w = self.add_it_up(self.additive_fitness_value_for_dip)
        self.assertEqual(w, 5.0)

    def testAllMult(self):
        self.assertEqual(len(self.all_mult), 5)
        w = self.prod_it_up(self.mult_fitness_value_for_dip)
        self.assertEqual(w, 1.0)

    def testAdditiveFitnessAggregator(self):
        w = self.agg_add_w(self.additive_fitness_value_for_dip)
        self.assertEqual(w, 1.0)

    def testMultFitnessAggregator(self):
        w = self.agg_mult_w(self.mult_fitness_value_for_dip)
        self.assertEqual(w, 1.0)

    def testSizeMismatchException(self):
        t = [fp11w.SlocusMult()] * 4  # no. call backs != no. loci
        tw = fp11m.MultiLocusGeneticValue(t)
        with self.assertRaises(ValueError):
            x = tw(self.pop.diploids[0], self.pop)

    def testPicklingAddFitness(self):
        p = pickle.dumps(self.agg_add_w)
        up = pickle.loads(p)
        self.assertEqual(type(self.agg_add_w),type(up))

    def testPicklingAddTrait(self):
        p = pickle.dumps(self.agg_add_t)
        up = pickle.loads(p)
        self.assertEqual(type(self.agg_add_t),type(up))

    def testPicklingMultFitness(self):
        p = pickle.dumps(self.agg_mult_w)
        up = pickle.loads(p)
        self.assertEqual(type(self.agg_mult_w),type(up))

    def testPicklingMultTrait(self):
        p = pickle.dumps(self.agg_mult_t)
        up = pickle.loads(p)
        self.assertEqual(type(self.agg_mult_t),type(up))

    def testPicklingMultiLocusGeneticValue(self):
        p = pickle.dumps(self.additive_w)
        up = pickle.loads(p)
        self.assertEqual(type(up),type(self.additive_w))

    def test_MultilocusGeneticValue_copy(self):
        try:
            f = fp11m.MultiLocusGeneticValue(self.additive_w.fitness_functions)
        except:
            self.fail("unexpected exception")

class testRecombination(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.rng = fp11.GSLrng(42)

    def testMakeBinomial(self):
        v = [0.5] * 5
        x = fp11m.binomial_rec(v)

    def testMakePoisson(self):
        v = [1e-3] * 5
        x = fp11m.poisson_rec(v)


    def testWarningsDeprecated(self):
        with self.assertWarns(DeprecationWarning):
            x = fp11m.binomial_rec(self.rng,1e-3)
        with self.assertWarns(DeprecationWarning):
            x = fp11m.binomial_rec(self.rng,[1e-3]*4)
        with self.assertWarns(DeprecationWarning):
            x = fp11m.poisson_rec(self.rng,1e-3)
        with self.assertWarns(DeprecationWarning):
            x = fp11m.poisson_rec(self.rng,[1e-3]*4)

if __name__ == "__main__":
    unittest.main()




import cppimport
cppimport.force_rebuild()
cppimport.set_quiet(False)
snowdrift = cppimport.imp("snowdrift")
test_polymorphism = cppimport.imp("test_polymorphism")
import unittest
import fwdpy11.fitness as fp11w
import re


class testFitnessPolymorphism(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.w = [
            fp11w.SlocusAdditive(),
            fp11w.SlocusMult(),
            snowdrift.SlocusSnowdrift(0.2, -0.2, 1, -2)]
        self.v = test_polymorphism.test_callback_names(self.w)

    def testCallbackNameLen(self):
        self.assertEqual(len(self.v), 3)

    def testReturnValuesSensible(self):
        p1 = re.compile('additive_diploid')
        p2 = re.compile('multiplicative_diploid')
        p3 = re.compile('snowdrift_diploid')
        self.assertEqual(len(p1.findall(self.v[0])), 1)
        self.assertEqual(len(p2.findall(self.v[1])), 1)
        self.assertEqual(len(p3.findall(self.v[2])), 1)

if __name__ == "__main__":
    unittest.main()




#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import unittest
import fwdpy11 as fp11
import fwdpy11.fitness as fp11w
from quick_pops import quick_nonneutral_slocus
import fwdpy11.python_genetic_values as pgv


def additive(dip, gametes, mutations):
    """
    Pure Python implementation of additive fitness.
    """
    rv = 0.0
    for i in gametes[dip.first].smutations:
        rv += mutations[i].s
    for i in gametes[dip.second].smutations:
        rv += mutations[i].s
    return max(0.0, 1 + rv)


class testCustomAdditive(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.pop = quick_nonneutral_slocus()
        self.acpp = fp11w.SlocusAdditive(2.0)
        self.a = pgv.GeneticValue(additive)

    def testDiploidFitnesses(self):
        """
        Compare our Python implementation to
        our C++ implementation.
        """
        for i in self.pop.diploids:
            self.assertEqual(self.acpp(i, self.pop), self.a(i, self.pop))


class testAdditiveEvolution(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from fwdpy11.ezparams import mslike
        from fwdpy11 import ExpS
        from fwdpy11.model_params import SlocusParams
        self.pop = fp11.SlocusPop(1000)
        self.pdict = mslike(
            self.pop, simlen=100, dfe=ExpS(0, 1, 1, -0.1),
            pneutral=0.95)
        self.pdict['gvalue'] = pgv.GeneticValue(additive)
        self.params = SlocusParams(**self.pdict)
        self.rng = fp11.GSLrng(42)

    def testValidate(self):
        try:
            self.params.validate()
        except:
            self.fail("Parameters failed to validate")

    def testEvolve(self):
        from fwdpy11.wright_fisher import evolve
        try:
            evolve(self.rng, self.pop, self.params)
        except:
            self.fail("unexpected exception")


if __name__ == "__main__":
    unittest.main()



# Tests of the fwdpy11.fitness module

import unittest
import fwdpy11


class testObject_repr(unittest.TestCase):
    def test_SlocusAdditive(self):
        from fwdpy11.fitness import SlocusAdditive
        w = SlocusAdditive(2.0)
        ww = eval('fwdpy11.' + repr(w))
        self.assertEqual(type(ww), SlocusAdditive)
        self.assertEqual(ww.scaling, 2.0)
        w = SlocusAdditive(1.0)
        ww = eval('fwdpy11.' + repr(w))
        self.assertEqual(type(ww), SlocusAdditive)
        self.assertEqual(ww.scaling, 1.0)

    def test_SlocusMult(self):
        from fwdpy11.fitness import SlocusMult
        w = SlocusMult(1.0)
        ww = eval('fwdpy11.' + repr(w))
        self.assertEqual(type(ww), SlocusMult)


if __name__ == "__main__":
    unittest.main()



# Test behavior of fixations in simulations.
# As of 0.1.3a1, fixations are to be sorted by position
# for all types of simulations, allowing for easier lookup.

import unittest
import os
import fwdpy11
import fwdpy11.util
import fwdpy11.wright_fisher_qtrait
import numpy as np
from quick_pops import quick_nonneutral_slocus
from quick_pops import quick_slocus_qtrait_pop_params
from quick_pops import quick_mlocus_qtrait_change_optimum


class testFixationsAreSorted(unittest.TestCase):
    def test_fixations_SlocusPop(self):
        from fwdpy11 import ExpS
        pop = quick_nonneutral_slocus(N=1000,
                                      simlen=10000,
                                      dfe=ExpS(0, 1, 1, 0.05))
        self.assertTrue(len(pop.fixations) > 0)
        self.assertEqual(len(pop.fixations), len(pop.fixation_times))
        fpos = [i.pos for i in pop.fixations]
        self.assertTrue(sorted(fpos))
        non_neutral_fixations = [
            i.key for i in pop.fixations if i.neutral is False]
        # If this test fails, we have sim parameters
        # that are not useful for testing:
        self.assertTrue(len(non_neutral_fixations) > 0)
        for ni in non_neutral_fixations:
            self.assertFalse(any(i.key == ni for i in pop.mutations))


class testFixationsAreSortedQtraitSim(unittest.TestCase):
    """
    In sims of quantitative traits, selected mutations are
    retained, as they still affect mean trait value, and thus
    mean distance from optimum, and thus mean fitness.
    However, they are entered into pop.fixations so that
    their fixation times may be recorded.
    """

    @unittest.skipIf("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true",
                     "Skipping this test on Travis CI.")
    def test_sorted_fixations_MlocusPopQtrait_with_pruning(self):
        pop = quick_mlocus_qtrait_change_optimum(N=1000, simlen=10000)
        self.assertTrue(len(pop.fixations) > 0)
        self.assertEqual(len(pop.fixations), len(pop.fixation_times))
        fpos = [i.pos for i in pop.fixations]
        self.assertTrue(sorted(fpos))
        non_neutral_fixations = [
            i.key for i in pop.fixations if i.neutral is False]
        # If this test fails, we have sim parameters
        # that are not useful for testing:
        self.assertTrue(len(non_neutral_fixations) > 0)
        for ni in non_neutral_fixations:
            # These non-neutral fixations
            # must still be found in the pop.
            self.assertTrue(any(i.key == ni for i in pop.mutations))
        neutral_fixations = [i.key for i in pop.fixations if i.neutral is True]
        self.assertTrue(len(neutral_fixations) > 0)
        for ni in neutral_fixations:
            self.assertTrue(any(i.key == ni for i in pop.mutations) is False)

    @unittest.skipIf("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true",
                     "Skipping this test on Travis CI.")
    def test_sorted_fixations_MlocusPopQtrait_without_pruning(self):
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            pop = quick_mlocus_qtrait_change_optimum(N=1000, simlen=10000, prune_selected = True)
            self.assertTrue(len(pop.fixations) > 0)
            self.assertEqual(len(pop.fixations), len(pop.fixation_times))
            fpos = [i.pos for i in pop.fixations]
            self.assertTrue(sorted(fpos))
            non_neutral_fixations = [
                i.key for i in pop.fixations if i.neutral is False]
            # If this test fails, we have sim parameters
            # that are not useful for testing:
            self.assertTrue(len(non_neutral_fixations) > 0)
            for ni in non_neutral_fixations:
                # These non-neutral fixations
                # must not be found in the pop.
                self.assertTrue(any(i.key == ni for i in pop.mutations) is False)
            neutral_fixations = [i.key for i in pop.fixations if i.neutral is True]
            self.assertTrue(len(neutral_fixations) > 0)
            for ni in neutral_fixations:
                self.assertTrue(any(i.key == ni for i in pop.mutations) is False)

class tests_MultipleFixationsAtPosition(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from fwdpy11.model_params import SlocusParamsQ
        self.pop, self.pdict = quick_slocus_qtrait_pop_params()
        self.params = SlocusParamsQ(**self.pdict)
        self.params.demography = np.array([self.pop.N], dtype=np.uint32)
        self.params.mutrate_n = 0.0  # disallow neutral mutations
        self.rng = fwdpy11.GSLrng(42)
        fwdpy11.util.add_mutation(
            self.rng, self.pop, 2 * self.pop.N, (0.1, 0.0, 0.0), 0)

    def test_two_neutral_fixations(self):
        # Evolve pop for 1 gen.
        # This will push mutation into fixations:
        fwdpy11.wright_fisher_qtrait.evolve(self.rng, self.pop, self.params)
        self.assertEqual(len(self.pop.fixations), 1)
        # Add back the exact same mutation:
        fwdpy11.util.add_mutation(
            self.rng, self.pop, 2 * self.pop.N, (0.1, 0.0, 0.0), 0)
        # Evolve again for 1 gen:
        fwdpy11.wright_fisher_qtrait.evolve(self.rng, self.pop, self.params)
        self.assertEqual(len(self.pop.fixations), 2)
        self.assertEqual(list(self.pop.fixation_times), [1, 2])


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

"""
    Define classes for parameterizing simulations.

    The classes take either kwargs or an expanded dict as arguments.
    The names of the kwargs are the same as the names of class properties.
    The properties themselves are gettable/settable in the usual way.

    Each class has a validate function that attempts to do a
    full error-checking on the data stored in a class instance.

    The module fwdpy11.ezparams contains some functions for rapidly setting up
    dictionaries of parameters for common cases.

    .. todo::
        The classes should make deep copies of the input so that data
        from the calling environment are not changed, which can happen
        in some cases.
        There is an issue for C++ types holding
        things like Python objects.  Currently,
        tests/test_python_genetic_values.py
        will segfault if we copy.deepcopy the params objects.
"""

import warnings


def _validate_types(data, typename, strict):
    for i in data:
        if (strict is True and type(i) is not typename) or \
           isinstance(i, typename) is False:
            raise TypeError("invalid type: " + str(typename))


class ModelParams(object):
    """
    Base class for simulation parameters

    .. versionadded:: 0.1.1
    """

    def __init__(self, **kwargs):
        self.__nregions = None
        self.__sregions = None
        self.__recregions = None
        self.__demography = None
        self.__prune_selected = True
        for key, value in kwargs.items():
            if key in dir(self) and key[:1] != "_":
                setattr(self, key, value)
            elif key not in dir(self):
                raise ValueError(key, " not a valid parameter for this class.")
    @property
    def nregions(self):
        """
        Get or set the neutral regions.
        The implementation of the setter
        is handled by derived classes.
        Look there for more details.
        """
        return self.__nregions

    @property
    def sregions(self):
        """
        Get or set the selected regions.
        The implementation of the setter
        is handled by derived classes.
        Look there for more details.
        """
        return self.__sregions

    @property
    def recregions(self):
        """
        Get or set the recombination regions.
        The implementation of the setter
        is handled by derived classes.
        Look there for more details.
        """
        return self.__recregions

    @property
    def prune_selected(self):
        """
        Get or set whether or not selected fixations
        are to be removed during simulaton.

        When setting, a bool is required.
        """
        return self.__prune_selected

    @prune_selected.setter
    def prune_selected(self, value):
        self.__prune_selected = bool(value)

    @nregions.setter
    def nregions(self, nregions):
        self.__nregions = nregions

    @sregions.setter
    def sregions(self, sregions):
        self.__sregions = sregions

    @recregions.setter
    def recregions(self, recregions):
        self.__recregions = recregions

    @property
    def demography(self):
        """
        Get or set demographic history.
        The setters are fully implemented in
        derived classes.  Look there for
        type info, as the details are model-dependent.
        """
        return self.__demography

    @demography.setter
    def demography(self, value):
        self.__demography = value

    def validate(self):
        """
        Error check model params.

        :raises ValueError: Throws ValueError if validation fails.

        .. note:: Demography objects must be validated in derived classes.
        """
        if self.nregions is None:
            raise ValueError("neutral regions cannot be None")
        if self.sregions is None:
            raise ValueError("selected regions cannot be None")
        if self.recregions is None:
            raise ValueError("recombination regions cannot be None")
        if self.demography is None:
            raise ValueError("demography cannot be None")
        if self.prune_selected is None:
            raise ValueError("prune_selected cannot be None")


def _validate_single_deme_demography(value):
    import numpy as np
    if any(i < 0 for i in value):
        raise ValueError("all population sizes must be > 0")
    if (type(value) is np.ndarray) is False:
        raise ValueError("Type for population size " +
                         "history must be numpy.array")


def _validate_single_locus_rates(data):
    for key, value in data.items():
        if value is None:
            raise ValueError(key + " cannot be None")
        if value < 0.:
            raise ValueError("All mutation and recombination " +
                             "rates must be non-negative.")


class SlocusParams(ModelParams):
    """
    Simulation parameters for single-locus, single-deme simulations.

    .. note::
        The demography property for this class is a numpy
        array with dtype = np.uint32.

        When setting nregions or recregions, lists of
        instances of :class:`fwdpy11.Region` are
        required.

        When setting sregions, lists of instances of
        :class:`fwdpy11.Sregion` are required.

    .. note::
        If no gvalue is assigned, then
        :class:`fwdpy11.fitness.SlocusMult` with
        a scaling of 2.0 will be used.

    .. versionadded:: 0.1.1
    """

    def __init__(self, **kwargs):
        self.__mutrate_n = None
        self.__mutrate_s = None
        self.__recrate = None
        self.__gvalue = None
        self.__pself = 0.0
        gv_present = False
        if 'gvalue' in kwargs:
            gv_present = True
        super(SlocusParams, self).__init__(**kwargs)
        if gv_present is False:
            from fwdpy11.fitness import SlocusMult
            self.gvalue = SlocusMult(2.0)

    def __set_rates(self, input_rates):
        try:
            if len(input_rates) != 3:
                raise ValueError("len(input_rates) != 3")
            self.mutrate_n = input_rates[0]
            self.mutrate_s = input_rates[1]
            self.recrate = input_rates[2]
            _validate_single_locus_rates(self.rates)
        except:
            try:
                if isinstance(input_rates, dict):
                    for key, value in input_rates.items():
                        if key == "mutrate_n":
                            setattr(self, key, float(value))
                        elif key == "mutrate_s":
                            setattr(self, key, float(value))
                        elif key == "recrate":
                            setattr(self, key, float(value))
                        else:
                            raise ValueError("invalid key/value " +
                                             "pair for setting rates: ",
                                             key, ', ' + str(value))
                else:
                    raise
            except:
                raise
        _validate_single_locus_rates(self.rates)

    @property
    def mutrate_n(self):
        """
        Get or set the neutral mutation rate.
        When setting, a float >= 0.0 is required.
        """
        return self.__mutrate_n

    @mutrate_n.setter
    def mutrate_n(self, neutral_mutation_rate):
        try:
            float(neutral_mutation_rate)
        except:
            raise ValueError("neutral mutation rate must be a float")
        if neutral_mutation_rate < 0.0:
            raise ValueError("neutral_mutation_rate must be >= 0.0")
        self.__mutrate_n = neutral_mutation_rate

    @property
    def mutrate_s(self):
        """
        Get or set the selected mutation rate.
        When setting, a float >= 0.0 is required.
        """
        return self.__mutrate_s

    @mutrate_s.setter
    def mutrate_s(self, selected_mutation_rate):
        try:
            float(selected_mutation_rate)
        except:
            raise ValueError("selected mutation rate must be a float")
        if selected_mutation_rate < 0.0:
            raise ValueError("selected mutation rate must be >= 0.0")
        self.__mutrate_s = selected_mutation_rate

    @property
    def recrate(self):
        """
        Get or set the recombination rate.
        When setting, a float >= 0.0 is required.
        """
        return self.__recrate

    @recrate.setter
    def recrate(self, recombination_rate):
        try:
            float(recombination_rate)
        except:
            raise ValueError("recombination rate must be a float")
        if recombination_rate < 0.0:
            raise ValueError("recombination rate must be >= 0.0")
        self.__recrate = recombination_rate

    @property
    def pself(self):
        """
        Get or set selfing probability.
        When setting, a float in the interval
        [0,1] is required.
        """
        return self.__pself

    @pself.setter
    def pself(self, prob_selfing):
        try:
            float(prob_selfing)
        except:
            raise ValueError("selfing probability must be a float")
        if prob_selfing < 0.0 or prob_selfing > 1.0:
            raise ValueError(
                "invalid selfing probability: " + str(prob_selfing))
        self.__pself = float(prob_selfing)

    @property
    def rates(self):
        """
        Get or set all rates at once.
        When getting, a dict is returned.
        When setting, a list or tuple of the
        three rates (neutral, seleted, recombination)
        may be used, or a dict with the keys mutrate_n,
        mutrate_s, and recrate.
        """
        return {'mutrate_n': self.__mutrate_n,
                'mutrate_s': self.__mutrate_s,
                'recrate': self.__recrate}

    @rates.setter
    def rates(self, input_rates):
        self.__set_rates(input_rates)

    @property
    def gvalue(self):
        """
        Get or set genetic value calculator.
        When setting, an instance of
        :class:`fwdpy11.fitness.SlocusFitness`
        is required.
        """
        return self.__gvalue

    @gvalue.setter
    def gvalue(self, f):
        from fwdpy11.fitness import SlocusFitness
        if isinstance(f, SlocusFitness) is False:
            raise ValueError("invalid genetic value type: " + str(type(f)))
        self.__gvalue = f

    @ModelParams.demography.setter
    def demography(self, demog):
        _validate_single_deme_demography(demog)
        ModelParams.demography.fset(self, demog)

    @ModelParams.nregions.setter
    def nregions(self, neutral_regions):
        from fwdpy11 import Region
        _validate_types(neutral_regions, Region, True)
        ModelParams.nregions.fset(self, neutral_regions)

    @ModelParams.recregions.setter
    def recregions(self, recombination_regions):
        from fwdpy11 import Region
        _validate_types(recombination_regions, Region, True)
        ModelParams.recregions.fset(self, recombination_regions)

    @ModelParams.sregions.setter
    def sregions(self, selected_regions):
        from fwdpy11 import Sregion
        _validate_types(selected_regions, Sregion, False)
        ModelParams.sregions.fset(self, selected_regions)

    def validate(self):
        """
        Sanity-check parameters.
        """
        super(SlocusParams, self).validate()
        from fwdpy11.fitness import SlocusFitness

        _validate_single_locus_rates(self.rates)

        if callable(self.gvalue) is False:
            raise ValueError("genetic value function must be callable")

        # If individual rates are nonzero, then
        # corresponding lists of regions cannot be empty:
        checks = {'mutrate_n': self.nregions,
                  'mutrate_s': self.sregions,
                  'recrate': self.recregions}
        for key, value in self.rates.items():
            if value > 0.0 and len(checks[key]) == 0:
                raise ValueError(key + ' = ' + str(value) +
                                 ' is associated with empty list of regions.')
            # Issue warning as this may reflect a mistake:
            if value == 0. and len(checks[key]) > 0:
                warnings.warn(key + ' = ' + str(value) +
                              ' has nonzero region length ('
                              + str(checks[key]) +
                              ').  Is this intentional?')

        if isinstance(self.gvalue, SlocusFitness) is False:
            raise ValueError("invalid genetic value type: " +
                             type(self.__gvalue_data['gvalue']))

        _validate_single_deme_demography(self.demography)


class SlocusParamsQ(SlocusParams):
    """
    Single locus parameter object for quantitative trait simulations

    .. note::
        The demography property for this class is a numpy
        array with dtype = np.uint32.

        When setting nregions or recregions, lists of
        instances of :class:`fwdpy11.Region` are
        required.

        When setting sregions, lists of instances of
        :class:`fwdpy11.Sregion` are required.

    .. note::
        If no gvalue is assigned, then
        :class:`fwdpy11.trait_values.SlocusAdditiveTrait` with
        a scaling of 2.0 will be used.

        If no trait2w is assigned, then
        :class:`fwdpy11.wright_fisher_qtrait.GSS` with VS = 1.0
        and O = 0.0 will be used.

    .. versionadded:: 0.1.1
    """

    def __init__(self, **kwargs):
        self.__trait_to_fitness = None
        self.__noise = None
        gv_present = False
        if 'gvalue' in kwargs:
            gv_present = True
        trait2w_present = False
        if 'trait2w' in kwargs or 'trait_to_fitness' in kwargs:
            trait2w_present = True

        super(SlocusParamsQ, self).__init__(**kwargs)
        if gv_present is False:
            from fwdpy11.trait_values import SlocusAdditiveTrait
            self.gvalue = SlocusAdditiveTrait(2.0)
        if trait2w_present is False:
            from fwdpy11.wright_fisher_qtrait import GSS
            self.trait2w = GSS(VS=1.0, O=0.0)

        if self.prune_selected is True:
            warnings.warn("setting prune_selected to False", UserWarning)
            self.prune_selected = False

    @property
    def trait2w(self):
        """
        Get or set the trait value to fitness mapping function.
        When setting, a callable is required.
        """
        return self.__trait_to_fitness

    @property
    def noise(self):
        """
        Get or set the noise function.
        When setting, a callable is required.
        """
        return self.__noise

    @trait2w.setter
    def trait2w(self, trait_to_fitness):
        if callable(trait_to_fitness) is False:
            raise ValueError("trait_to_fitness must be a callable.")
        self.__trait_to_fitness = trait_to_fitness

    @property
    def trait_to_fitness(self):
        warnings.warn("\'trait_to_fitness\' property is deprecated. "
                      "Please use \'trait2w\'",
                      DeprecationWarning)
        return self.trait2w

    @trait_to_fitness.setter
    def trait_to_fitness(self, value):
        warnings.warn("\'trait_to_fitness\' property is deprecated. "
                      "Please use \'trait2w\'",
                      DeprecationWarning)
        self.trait2w = value

    @noise.setter
    def noise(self, noise):
        if callable(noise) is False:
            raise ValueError("noise must be a callable.")
        self.__noise = noise

    def validate(self):
        super(SlocusParamsQ, self).validate()

        if callable(self.trait2w) is False:
            raise ValueError("trait to fitness mapper must be a callable")

        if self.noise is not None and callable(self.noise) is False:
            raise ValueError("noise function must be callable")

        if self.prune_selected is True:
            raise ValueError("invalid value for prune_selected: " +
                             str(self.prune_selected))

        if self.__trait_to_fitness is None:
            raise ValueError("trait to fitness mapping " +
                             "function cannot be None")


def _validate_multilocus_rates(data):
    for key, value in data.items():
        if value is None:
            raise ValueError(key + " cannot be None")
        if isinstance(value, list) is False:
            raise ValueError(key + " must be a list of float")
        for i in value:
            try:
                float(i)
            except:
                raise ValueError(
                    "rates must be floating point values")
        if any(i < 0. for i in value):
            raise ValueError("all mutation and recombination rates " +
                             "must be non-negative.")
    # All rate lists must be equal in length
    lengths = set([len(value) for key, value in data.items()])
    if len(lengths) > 1:
        raise ValueError("all lists of mutation and recombination " +
                         "rates must be equal length.")


def _sanity_check_mutlilocus_rates_and_regions(rate_data,
                                               nregions,
                                               sregions,
                                               recregions):
    for i, j in zip(rate_data['mutrates_n'], nregions):
        if i > 0.0 and len(j) == 0:
            raise ValueError(
                "non-zero rate found associated with empty neutral locus")
    for i, j in zip(rate_data['mutrates_s'], sregions):
        if i > 0.0 and len(j) == 0:
            raise ValueError(
                "non-zero rate found associated with empty selected locus")
    for i, j in zip(rate_data['recrates'], recregions):
        if i > 0.0 and len(j) == 0:
            raise ValueError(
                "non-zero rate found associated with empty "
                "recombination locus")


class MlocusParams(ModelParams):
    """
    Model parameters for multi-locus simulation

    .. note::
        The demography property for this class is a numpy
        array with dtype = np.uint32.

        When setting nregions or recregions, lists of
        lists are required.  The inner lists may be empty
        or they may contain instances of
        :class:`fwdpy11.Region`.

        Similarly, sregions inner lists should contain
        instances of :class:`fwdpy11.Sregion`.
        They may also be empty.

    .. note::
        If aggregator is not assigned, then
        :class:`fwdpy11.multilocus.AggMultFitness` will be
        used.

        If gvalue is not assigned during construction, then
        :class:`fwdpy11.multilocus.MultiLocusGeneticValue`
        will be used and filled with instances of
        :class:`fwdpy11.fitness.SlocusMult` with
        scaling set to 2.0.  This assignment to gvalue is
        only possible if sregions is assigned during construction,
        so that we know how many loci there are.

    .. versionadded:: 0.1.1
    """

    def __init__(self, **kwargs):
        self.__mutrates_n = None
        self.__mutrates_s = None
        self.__recrates = None
        self.__interlocus = None
        self.__gvalue = None
        self.__agg = None
        self.__pself = 0.0
        super(MlocusParams, self).__init__(**kwargs)
        if self.aggregator is None:
            from fwdpy11.multilocus import AggMultFitness
            self.aggregator = AggMultFitness()
        if self.gvalue is None:
            if self.sregions is not None:
                from fwdpy11.multilocus import MultiLocusGeneticValue
                from fwdpy11.fitness import SlocusMult
                self.gvalue = MultiLocusGeneticValue([SlocusMult(2.0)] *
                                                     len(self.sregions))

    @property
    def gvalue(self):
        """
        Get or set genetic value calculation object.
        When setting, either a list of
        :class:`fwdpy11.fitness.SlocusFitness` or
        and instance of
        :class:`fwdpy11.multilocus.MultiLocusGeneticValue`
        is required.
        """
        return self.__gvalue

    @gvalue.setter
    def gvalue(self, f):
        from fwdpy11.multilocus import MultiLocusGeneticValue
        from fwdpy11.fitness import SlocusFitness
        if isinstance(f, list) is True:
            if any(isinstance(i, SlocusFitness) for i in f) is False:
                raise ValueError("all elements in list must " +
                                 "be single-locus genetic value objects.")
            self.__gvalue = MultiLocusGeneticValue(f)
        elif isinstance(f, MultiLocusGeneticValue) is True:
            self.__gvalue = f
        else:
            raise ValueError("invalid genetic value type: " + str(type(f)))

    @property
    def agg(self):
        """
        Get or set the aggregator function.
        When setting, a callable is required.

        .. note::
            This is deprecated, use 'aggregator'
            instead.

        .. deprecated:: 0.13.0
        """
        warnings.warn("\'agg\' property is deprecated. "
                      "Please use \'aggregator\'",
                      DeprecationWarning)
        return self.aggregator

    @agg.setter
    def agg(self, aggregator):
        warnings.warn("\'agg\' property is deprecated. "
                      "Please use \'aggregator\'",
                      DeprecationWarning)
        self.aggregator = aggregator

    @property
    def aggregator(self):
        """
        Get or set the aggregator function.
        When setting, a callable is required.
        """
        return self.__agg

    @aggregator.setter
    def aggregator(self, aggregator):
        if callable(aggregator) is False:
            raise ValueError("aggregator must be a callable")
        self.__agg = aggregator

    @property
    def mutrates_n(self):
        """
        Get or set the neutral mutation rates.
        When setting a list of floats >= 0.0
        is required.
        """
        return self.__mutrates_n

    @mutrates_n.setter
    def mutrates_n(self, neutral_mutation_rates):
        try:
            for i in neutral_mutation_rates:
                float(i)
        except:
            raise ValueError("neutral mutation rates must all be float")

        if any(i < 0.0 for i in neutral_mutation_rates):
            raise ValueError("neutral mutation rates must be >= 0.0")

        self.__mutrates_n = neutral_mutation_rates

    @property
    def mutrates_s(self):
        """
        Get or set the selected mutation rates.
        When setting a list of floats >= 0.0
        is required.
        """
        return self.__mutrates_s

    @mutrates_s.setter
    def mutrates_s(self, selected_mutation_rates):
        try:
            for i in selected_mutation_rates:
                float(i)
        except:
            raise ValueError("selected mutation rates must all be float")

        if any(i < 0.0 for i in selected_mutation_rates):
            raise ValueError("selected mutation rates must be >= 0.0")

        self.__mutrates_s = selected_mutation_rates

    @property
    def recrates(self):
        """
        Get or set the recombination rates.
        When setting a list of floats >= 0.0
        is required.
        """
        return self.__recrates

    @recrates.setter
    def recrates(self, recombination_rates):
        try:
            for i in recombination_rates:
                float(i)
        except:
            raise ValueError("recombination rates must all be float")

        if any(i < 0.0 for i in recombination_rates):
            raise ValueError("recombination rates must be >= 0.0")

        self.__recrates = recombination_rates

    @ModelParams.demography.setter
    def demography(self, demog):
        _validate_single_deme_demography(demog)
        ModelParams.demography.fset(self, demog)

    @ModelParams.nregions.setter
    def nregions(self, neutral_regions):
        from fwdpy11 import Region
        for i in neutral_regions:
            _validate_types(i, Region, True)
        ModelParams.nregions.fset(self, neutral_regions)

    @ModelParams.recregions.setter
    def recregions(self, recombination_regions):
        from fwdpy11 import Region
        for i in recombination_regions:
            _validate_types(i, Region, True)
        ModelParams.recregions.fset(self, recombination_regions)

    @ModelParams.sregions.setter
    def sregions(self, selected_regions):
        from fwdpy11 import Sregion
        for i in selected_regions:
            _validate_types(i, Sregion, False)
        ModelParams.sregions.fset(self, selected_regions)

    @property
    def pself(self):
        """
        Get or set selfing probability.
        When setting, a float in the interval
        [0,1] is required.
        """
        return self.__pself

    @pself.setter
    def pself(self, prob_selfing):
        self.__pself = float(prob_selfing)

    @property
    def rates(self):
        """
        Get or set all rates at once.
        When getting, a dict is returned.
        When setting, a list or tuple of the
        three rates (neutral, seleted, recombination)
        may be used, or a dict with the keys mutrates_n,
        mutrates_s, and recrates.
        """
        return {'mutrates_n': self.__mutrates_n,
                'mutrates_s': self.__mutrates_s,
                'recrates': self.__recrates}

    @rates.setter
    def rates(self, rates):
        self.__set_rates(rates)

    def __set_rates(self, rates):
        try:
            self.__mutrates_n = rates[0]
            self.__mutrates_s = rates[1]
            self.__recrates = rates[2]
            _validate_multilocus_rates(self.rates)
        except:
            try:
                if isinstance(rates, dict):
                    for key, value in rates.items():
                        if key in self.__expected_mutrec_kwargs:
                            self.__mutrec_data[key] = value
                        else:
                            raise ValueError("invalid key/value pair " +
                                             "for setting rates: ",
                                             key, ', ' + str(value))
                else:
                    raise
            except:
                raise
        _validate_multilocus_rates(self.__mutrec_data)

    @property
    def interlocus(self):
        """
        Get or set interlocus recombination rates.
        When setting, a list of instances of
        :class:`fwdpy11.multilocus.InterlocusRecombination`
        is required.
        """
        return self.__interlocus

    @interlocus.setter
    def interlocus(self, interlocus_rec):
        from fwdpy11.multilocus import InterlocusRecombination as IR
        if any(isinstance(i, IR) for i in interlocus_rec) is False:
            raise ValueError("interlocus recombination objects must be "
                             "instances of "
                             "fwdpy11.multilocus.InterlocusRecombination")
        self.__interlocus = interlocus_rec

    def validate(self):
        """
        Sanity-check the class data
        """
        super(MlocusParams, self).validate()

        _validate_multilocus_rates(self.rates)

        if callable(self.aggregator) is False:
            raise ValueError("aggregator must be callable")

        # Lengths of all regions must be the same
        rlens = set((len(self.nregions),
                     len(self.sregions), len(self.recregions)))
        if len(rlens) > 1:
            raise ValueError("length of all region " +
                             "containers must be equal: " +
                             ','.join([str(i) for i in rlens]))

        # Lengths of all lists of rates must be the same
        ratelens = set([len(value) for key, value in self.rates.items()])
        if len(ratelens) > 1:
            raise ValueError("length of per-locus " +
                             "rate containers must be equal: " +
                             ','.join([str(i) for i in rlens]))

        if rlens != ratelens:
            raise ValueError("region and rate containers of different lengths")

        _sanity_check_mutlilocus_rates_and_regions(self.rates,
                                                   self.nregions,
                                                   self.sregions,
                                                   self.recregions)

        # For k loci, there must be k-1 interlocus-recombination rates
        nloci = rlens.pop()
        if len(self.interlocus) + 1 != nloci:
            raise ValueError("invalid parameters: " +
                             str(nloci) + " loci, but only " +
                             str(len(self.interlocus)) +
                             " between-locus recombination functions.")

        if len(self.gvalue) != nloci:
            raise ValueError("Number of loci does not equal " +
                             "number of functions in genetic value object: " +
                             str(nloci) + " != " + str(len(self.aggregator)))


class MlocusParamsQ(MlocusParams):
    """
    Multi-locus model parameters for quantitative trait simulations

    .. note::
        The demography property for this class is a numpy
        array with dtype = np.uint32.

    .. note::
        If aggregator is not assigned, then
        :class:`fwdpy11.multilocus.AggMultTrait` will be
        used.

        If gvalue is not assigned during construction, then
        :class:`fwdpy11.multilocus.MultiLocusGeneticValue`
        will be used and filled with instances of
        :class:`fwdpy11.trait_values.SlocusAdditiveTrait` with
        scaling set to 2.0.  This assignment to gvalue is
        only possible if sregions is assigned during construction,
        so that we know how many loci there are.

        If trait2w is not set during construction, then
        :class:`fwdpy11.wright_fisher_qtrait.GSS` will be
        used with VS = 1.0 and O = 0.0.

    .. versionadded:: 0.1.1
    """

    def __init__(self, **kwargs):
        self.__trait2w = None
        self.__noise = None
        agg_present = False
        if 'agg' in kwargs or 'aggregator' in kwargs:
            agg_present = True
        gvalue_present = False
        if 'gvalue' in kwargs:
            gvalue_present = True
        trait2w_present = False
        if 'trait2w' in kwargs or 'trait_to_fitness' in kwargs:
            trait2w_present = True
        super(MlocusParamsQ, self).__init__(**kwargs)
        if agg_present is False:
            from fwdpy11.multilocus import AggMultTrait
            self.aggregator = AggMultTrait()
        if gvalue_present is False and self.sregions is not None:
            from fwdpy11.multilocus import MultiLocusGeneticValue
            from fwdpy11.trait_values import SlocusAdditiveTrait
            self.gvalue = MultiLocusGeneticValue([SlocusAdditiveTrait(2.0)] *
                                                 len(self.sregions))
        if trait2w_present is False:
            from fwdpy11.wright_fisher_qtrait import GSS
            self.trait2w = GSS(VS=1.0, O=0.0)

    @property
    def trait2w(self):
        """
        Get or set the trait value to fitness mapping function.
        When setting, a callable is required.
        """
        return self.__trait2w

    @trait2w.setter
    def trait2w(self, trait2w):
        if callable(trait2w) is False:
            raise ValueError("trait2w must be a callable.")
        self.__trait2w = trait2w

    @property
    def trait_to_fitness(self):
        """
        Get or set the trait value to fitness mapping function.
        When setting, a callable is required.

        .. deprecated:: 0.13.0
            Use trait2w instead.
        """
        warnings.warn("\'trait_to_fitness\' property is deprecated. "
                      "Please use \'trait2w\'",
                      DeprecationWarning)
        return self.trait2w

    @trait_to_fitness.setter
    def trait_to_fitness(self, value):
        warnings.warn("\'trait_to_fitness\' property is deprecated. "
                      "Please use \'trait2w\'",
                      DeprecationWarning)
        self.trait2w = value

    @property
    def noise(self):
        """
        Get or set the noise function.
        When setting, a callable is required.
        """
        return self.__noise

    @noise.setter
    def noise(self, noise):
        if callable(noise) is False:
            raise ValueError("noise must be a callable.")
        self.__noise = noise

    def validate(self):
        """
        Sanity-check class data
        """
        if callable(self.trait2w) is False:
            raise ValueError("trait2w must be a callable.")

        if self.noise is None:
            warnings.warn("noise parameter is None.  Defaults will be used.",
                          UserWarning)

        super(MlocusParamsQ, self).validate()



import unittest
from fwdpy11 import GSLrng
from fwdpy11.wright_fisher_qtrait import evolve, GaussianNoise, GSS
from fwdpy11.trait_values import SlocusAdditiveTrait
from fwdpy11.multilocus import MultiLocusGeneticValue
from quick_pops import quick_mlocus_qtrait_pop_params


class GaussianNoiseUpdater(GaussianNoise):
    """
    This noise updater inherits from the
    standard fwdpy11 object.  It defines an
    updater function to record how many
    times it has been called.
    """

    def __init__(self, rng, sd, mean=0.0):
        self.ncalls = 0
        super(GaussianNoiseUpdater, self).__init__(rng, sd, mean)

    def update(self, generation):
        self.ncalls += 1


class GSSupdater(GSS):
    def __init__(self, VS, O):
        super(GSSupdater, self).__init__(VS, O)
        self.ncalls = 0

    def update(self, pop):
        self.ncalls += 1


class testNoiseUpdater(unittest.TestCase):
    def test_noise_updater(self):
        from fwdpy11.model_params import MlocusParamsQ
        rng = GSLrng(42)
        pop, pdict = quick_mlocus_qtrait_pop_params()
        self.params = MlocusParamsQ(**pdict)
        n = GaussianNoiseUpdater(rng, 0.1)
        trait2w = GSSupdater(1.0, 0)
        self.params.noise = n
        self.params.trait2w = trait2w
        evolve(rng, pop, self.params)
        self.assertEqual(n.ncalls, pop.generation)
        self.assertEqual(trait2w.ncalls, pop.generation)


if __name__ == "__main__":
    unittest.main()



# Tests of types in fwdpy11.trait_values

import unittest
import fwdpy11


class testObject_repr(unittest.TestCase):
    def test_SlocusAdditiveTrait(self):
        from fwdpy11.trait_values import SlocusAdditiveTrait
        w = SlocusAdditiveTrait(1.0)
        ww = eval('fwdpy11.' + repr(w))
        self.assertEqual(type(ww), SlocusAdditiveTrait)
        self.assertEqual(ww.scaling, 1.0)

    def test_SlocusMultTrait(self):
        from fwdpy11.trait_values import SlocusMultTrait
        w = SlocusMultTrait(1.0)
        ww = eval('fwdpy11.' + repr(w))
        self.assertEqual(type(ww), SlocusMultTrait)
        self.assertEqual(ww.scaling, 1.0)

    def test_GBR(self):
        from fwdpy11.trait_values import SlocusGBRTrait
        w = SlocusGBRTrait()
        ww = eval('fwdpy11.' + repr(w))
        self.assertEqual(type(ww), SlocusGBRTrait)


if __name__ == "__main__":
    unittest.main()
